{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Homework 5\n",
    "## Due Tuesday, November 9th 2021 at 11:59 PM.\n",
    "\n",
    "[Problem 0](#p0).  Homework Workflow [10 pts]\n",
    "\n",
    "[Problem 1](#p1).  `@property` Decorator [35 pts]\n",
    "\n",
    "[Problem 2](#p3).  Binary Search Tree [55 pts]\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "**We will be using an autograder for this homework. Please don't change the specified function signature.**"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "<a id='p0'></a>\n",
    "## Problem 0: Homework Workflow\n",
    "Once you receive HW4 feedback, you will need to merge your `HW4-dev` branch into `master`. \n",
    "\n",
    "You will earn points for following all stages of the git workflow which involves:\n",
    "* 3pts for merging `HW4-dev` into `master`\n",
    "* 5pts for completing HW5 on `HW5-dev`\n",
    "* 2pts for making a PR on `HW5-dev` to merge into `master`\n",
    "![Sample Github Submission](../fig/HW5_submission.png)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "<a id='p1'></a>\n",
    "## Problem 1: `@property` Decorator\n",
    "\n",
    "Suppose we have a simple `Animal` class, which is initialized by its species:\n",
    "\n",
    "```python\n",
    "class Animal:\n",
    "    \n",
    "    def __init__(self, name, species):\n",
    "        self.name = name\n",
    "        self.species = species\n",
    "        \n",
    "    def __repr__(self):\n",
    "        return f'{self.name} ({self.species})'\n",
    "```\n",
    "\n",
    "We have clients who build applications on top of our `Animal` class and they want to directly modify the `species` attribute. For example, they may implement functions to transform animals.\n",
    "\n",
    "```python\n",
    "def transform(animal, into):\n",
    "    animal.species = into\n",
    "    return animal\n",
    "```\n",
    "\n",
    "However, one day it occurred to us that our `Animal` class is designed to be used in an [H. C. Andersen universe](https://en.wikipedia.org/wiki/Hans_Christian_Andersen#Fairy_tales_and_poetry). We really don't want people to make an `Animal` instance into a [Cthulhu](https://en.wikipedia.org/wiki/Cthulhu) [Cthulhu is not part of the H. C. Anderson universe]. How should we enforce this restriction? The obvious solution is to create a method to set the species, which checks if the new species is valid inside the method:\n",
    "\n",
    "```python\n",
    "class Animal:\n",
    "    \n",
    "    # a class attribute of the valid species in our universe\n",
    "    valid_species = {\n",
    "        'cat',\n",
    "        'dog',\n",
    "        'duck',\n",
    "        'elf',\n",
    "        'goblin',\n",
    "        'horse',\n",
    "        'human',\n",
    "        'mermaid',\n",
    "        'nightingale',\n",
    "        'pig',\n",
    "        'swan',\n",
    "        'wolf'\n",
    "    }\n",
    "    \n",
    "    def __init__(self, species):\n",
    "        self.species = species\n",
    "            \n",
    "    def __repr__(self):\n",
    "        return f'{self.name} ({self.species})'\n",
    "    \n",
    "    def set_species(self, into):\n",
    "        assert into in Animal.valid_species, Exception(f'invalid species: {into}')\n",
    "        self.species = into\n",
    "```\n",
    "\n",
    "There are two problems with this implementation: \n",
    "1. If our clients don't know they have to change their implementation of `transform` (use `set_species` instead of directly assign the `species` attribute), people can still do things like `a = transform(ugly_duckling, into='Cthulhu')`\n",
    "2. We really don't want to bother our clients to change their code because they may have used `animal.species = something` hundreds of times throughout their code.\n",
    "\n",
    "Fortunately, we have the `property` decorator. Read more about it [here](https://docs.python.org/3/library/functions.html#property). \n",
    "\n",
    "For this problem, you will utilize the `property` decorator to limit what species users can set their `Animal` object to. Please copy the following code into the file `animal.py` (please create the file and copy below code).\n",
    "\n",
    "```python\n",
    "class Animal:\n",
    "    \n",
    "    # a class attribute of the valid species in our universe\n",
    "    valid_species = {\n",
    "        'cat',\n",
    "        'dog',\n",
    "        'duck',\n",
    "        'elf',\n",
    "        'goblin',\n",
    "        'horse',\n",
    "        'human',\n",
    "        'mermaid',\n",
    "        'nightingale',\n",
    "        'pig',\n",
    "        'swan',\n",
    "        'wolf'\n",
    "    }\n",
    "    \n",
    "    def __init__(self, name, species):\n",
    "        self.name = name\n",
    "        self._species = species\n",
    "    \n",
    "    def __repr__(self):\n",
    "        return f'{self.name} ({self._species})'\n",
    "```\n",
    "\n",
    "\n",
    "### Part A [15 points]:\n",
    "\n",
    "We have also now designated the `species` attribute as \"non-user-facing\" by calling it `_species` because we do not want the user to access it directly. However, this presents a problem because in our first version of the `Animal` class, a client could access the species through the `species` attribute. You can check that the attributes of the `Animal` class are `name` and `_species` using the following steps (on Python shell):\n",
    "```python\n",
    ">>> dog = Animal('Fido', 'dog')\n",
    ">>> vars(dog)\n",
    "{'name': 'Fido',  '_species' : 'dog'}\n",
    "```\n",
    "\n",
    "To prevent the client's previous code from breaking, modify the `Animal` class in `animal.py` by adding a `species(self)` method to the `Animal` class. The `species` method should return the `_species` attribute of this `Animal`. Apply the [`property` decorator](https://docs.python.org/3/library/functions.html#property) to this method.\n",
    "\n",
    "If this is done correctly, then a client should still be able to retrieve the species of an `Animal` object as follows in the Python shell (you should use `print()` in an editor to print out):\n",
    "```python\n",
    ">>> dog = Animal('Fido', 'dog')\n",
    ">>> dog.species\n",
    "'dog'\n",
    "```\n",
    "\n",
    "### Part B [20 points]:\n",
    "\n",
    "What happens when the client tries to set the value of the `species` attribute as in the following:\n",
    "```python\n",
    ">>> dog.species = 'cat'\n",
    "```\n",
    "You don't need to write the answer to this question, but you should test it.\n",
    "\n",
    "Thus to prevent the client's previous code from breaking, we need to create another method that:\n",
    "1. Allows the client to set the species of an `Animal` object using `.species`\n",
    "2. And checks that the species attribute is set to a species in our `valid_species` list.\n",
    "\n",
    "Add another method `species(self, into)` to the `Animal` class. This method should set the `_species` attribute of this `Animal` to `into`, but before that, it makes sure that `into` is in the `valid_species` list. If `into` is not in the `valid_species` list, raise an `Exception`. Apply the `species.setter` decorator to this method.\n",
    "\n",
    "If this is done correctly, then a client should still be able to set the species of an `Animal` object (within the restricted list) as follows, in a Python shell (in an editor, you should use `print()` to print out):\n",
    "```python\n",
    ">>> dog = Animal('Fido', 'dog')\n",
    ">>> dog.species\n",
    "'dog'\n",
    ">>> dog.species = 'cat'\n",
    ">>> dog.species\n",
    "'cat'\n",
    ">>> dog.species = 'TheThing'\n",
    "...\n",
    "AssertionError: invalid species: TheThing\n",
    "```\n",
    "\n",
    "Do our clients need to change the implementation of their `transform` function now? \n",
    "Nope, and hopefully they give you a bonus!\n",
    "\n",
    "#### Deliverables\n",
    "- `animal.py`"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "<a id='p3'></a>\n",
    "## Problem 2: Binary Search Tree\n",
    "\n",
    "A binary search tree (BST) is a binary tree in which:\n",
    "* The left subtree of a node contains only nodes with keys less than the node’s key.\n",
    "* The right subtree of a node contains only nodes with keys greater than the node’s key.\n",
    "* The left and right subtree each must also be a binary search tree.\n",
    "\n",
    "The code below is a partially implemented binary search tree. `BSTNode` is a node in the binary search tree. It contains a key-value pair and has references to its child nodes. `BSTTable` is the binary search tree itself. It holds a reference to the root node of the tree and contains methods that manipulate the nodes. Each child of the root node and its child should be a `BSTNode`. A `BSTTable` can be used as a table to store key-value pairs. Our implementation will only be able to handle the case when the keys are comparable to each other. Simple client code of `BSTTable` looks like this:\n",
    "\n",
    "```python\n",
    "greektoroman = BSTTable()\n",
    "greektoroman.put('Athena',    'Minerva')\n",
    "greektoroman.put('Eros',      'Cupid')\n",
    "greektoroman.put('Aphrodite', 'Venus')\n",
    "greektoroman.get('Eros')\n",
    "```\n",
    "and will return\n",
    "```\n",
    "'Cupid'\n",
    "```\n",
    "\n",
    "In this problem, we will implement the `get` and `put` method of `BSTTable`. \n",
    "\n",
    "We start by copying the template code below into the file `binary_search_tree.py`.\n",
    "\n",
    "**Hint:** You can `print` a `BSTTable` or `BSTNode` to see its structure. It's not pretty, but should be enough for debugging purposes. If you get lost, please try to draw the schematic on a paper step-by-step for each node creation and try to understand what your implemented functions do. For example, \n",
    "\n",
    "```python\n",
    "print(greektoroman) \n",
    "```\n",
    "\n",
    "The above statement will print out: \n",
    "```\n",
    "BSTNode(Athena, Minerva)\n",
    "|\n",
    "|-(L)->BSTNode(Aphrodite, Venus)\n",
    "|      |\n",
    "|      |-(L)->None\n",
    "|      |\n",
    "|      |-(R)->None\n",
    "|\n",
    "|-(R)->BSTNode(Eros,Cupid)\n",
    "|      |\n",
    "|      |-(L)->None\n",
    "|      |\n",
    "|      |-(R)->None\n",
    "```\n",
    "\n",
    "**Note:** You may assume that the keys are unique.\n",
    "\n",
    "```python\n",
    "class BSTNode:\n",
    "\n",
    "    def __init__(self, key, val):\n",
    "        self.key, self.val = key, val\n",
    "        self.left = None\n",
    "        self.right = None\n",
    "        self.size = 1\n",
    "    \n",
    "    def __str__(self):\n",
    "        return f'BSTNode({self.key}, {self.val})' + \\\n",
    "               '\\n|\\n|-(L)->' + '\\n|      '.join(str(self.left ).split('\\n')) + \\\n",
    "               '\\n|\\n|-(R)->' + '\\n|      '.join(str(self.right).split('\\n'))\n",
    "        \n",
    "        \n",
    "class BSTTable:\n",
    "    def __init__(self):\n",
    "        self._root = None\n",
    "    \n",
    "    def __str__(self):\n",
    "        return str(self._root)\n",
    "    \n",
    "    def __len__(self):\n",
    "        return self._size(self._root)\n",
    "        \n",
    "    def put(self, key, val):\n",
    "        self._root = self._put(self._root, key, val)\n",
    "    \n",
    "    def get(self, key):\n",
    "        return self._get(self._root, key)\n",
    "    \n",
    "    def _put(self, node, key, val):\n",
    "        pass # TODO\n",
    "    \n",
    "    def _get(self, node, key):\n",
    "        pass # TODO\n",
    "    \n",
    "    @staticmethod\n",
    "    def _size(node):\n",
    "        return node.size if node else 0\n",
    "```\n",
    "\n",
    "### Part A [30 points]: \n",
    "\n",
    "The `put` method inserts a key-value pair into the `BSTTable`. It delegates the function call to `_put`. Please implement the non-public method `_put`. It takes a `node`, a `key`, and a `val` as arguments. `node` is the root of the subtree into which the key-value pair is to be inserted. After the insertion, the root of this subtree may change. `_put` should return the new root of this subtree. The difference of the input subtree and the returned subtree is that in the returned subtree, a new key-value pair has been inserted. If the key was already in the subtree, then we update the value of the corresponding node.\n",
    "\n",
    "**Note:** We assume all the keys are of the same type and are comparable to each other with `>`, `<` and `==`. So if all your keys are strings (e.g. `greektoroman`), they will be ordered lexicographically. If all your keys are numbers, they will be ordered from the smallest to the largest. \n",
    "\n",
    "**Hints:** \n",
    "* This is recursive.\n",
    "* Think about the base case: what should be returned if the `node` passed in is `None`? \n",
    "* Which subtree (left/right) should we insert if the `node` is not `None`? \n",
    "* Remember to update the `size` of the nodes.\n",
    "    - The `size` attribute of each `BSTNode` is the size of this subtree, i.e. the number of nodes including itself and all its descendants.\n",
    "\n",
    "### Part B [25 points]: \n",
    "\n",
    "The `get` method returns the value with the corresponding key. It delegates the function call to `_get`. Please implement the non-public method `_get`. It takes a `node` and a `key` as an argument. It will search for a node with the same key under the subtree whose root is `node`. If there is such a node, return the `val` of this node. `_get` should raise a `KeyError` if no such node is found.\n",
    "\n",
    "#### Deliverables\n",
    "- `binary_search_tree.py`"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.6.8"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
